#include "spmv.h"

#if DOUBLEPRECISION
texture<int2, 1> texRef;
#else
texture<REAL, 1> texRef;
#endif

/*----------------------------------------------------------*/
__global__ 
void csr_s_k(int n, int *d_ia, int *d_ja, REAL *d_a, REAL *d_y) {
/*----------------------------------------------------------*
 *     spmv CSR-scalar-kernel
 *----------------------------------------------------------*/
  int idx = blockIdx.x*blockDim.x+threadIdx.x;
  int thread_num = gridDim.x * blockDim.x;

  while (idx < n)
  {
    int p = d_ia[idx];
    int q = d_ia[idx+1];
    REAL r = 0.0;
    int i;
    for (i=p; i<q; i++)
    {
#if DOUBLEPRECISION
    int2 t = tex1Dfetch(texRef, d_ja[i-1]-1);
    r += d_a[i-1]*__hiloint2double(t.y, t.x);
#else
    r += d_a[i-1]*tex1Dfetch(texRef, d_ja[i-1]-1);
#endif
    }
    d_y[idx] = r;
    idx += thread_num;
  }
}

/*--------------------------------------------------------*/
void spmv_csr_scalar(struct csr_t *csr, REAL *x, REAL *y) {
  int *d_ia, *d_ja, i;
  REAL *d_a, *d_x, *d_y;
  int n = csr->n;
  int nnz = csr->nnz; 
  double t1,t2;
/*-------------------------- allocate Device Memory */
  cudaMalloc((void **)&d_ia, (n+1)*sizeof(int));
  cudaMalloc((void **)&d_ja, nnz*sizeof(int));
  cudaMalloc((void **)&d_a, nnz*sizeof(REAL));
  cudaMalloc((void **)&d_x, n*sizeof(REAL));
  cudaMalloc((void **)&d_y, n*sizeof(REAL));
/*--------------------- Memcpy */
  cudaMemcpy(d_ia, csr->ia, (n+1)*sizeof(int),
  cudaMemcpyHostToDevice);
  cudaMemcpy(d_ja, csr->ja, nnz*sizeof(int),
  cudaMemcpyHostToDevice);
  cudaMemcpy(d_a, csr->a, nnz*sizeof(REAL),
  cudaMemcpyHostToDevice);
  cudaMemcpy(d_x, x, n*sizeof(REAL),
  cudaMemcpyHostToDevice);
/*--------- texture binding */
  size_t offset;
#if DOUBLEPRECISION
  cudaBindTexture(&offset, texRef, d_x, n*sizeof(int2));
#else
  cudaBindTexture(&offset, texRef, d_x, n*sizeof(float));
#endif
  assert(offset == 0);
/*-------- set spmv kernel */
  int nthreads = min(MAXTHREADS, n);
  int gDim = (nthreads+BLOCKDIM-1)/BLOCKDIM;
  int bDim = BLOCKDIM;
/*-------- start timing */
  t1 = wall_timer();
  for (i=0; i<REPEAT; i++) {
    //cudaMemset((void *)d_y, 0, n*sizeof(REAL));
    csr_s_k<<<gDim, bDim>>>(n, d_ia, d_ja, d_a, d_y);
  }
/*-------- barrier for GPU calls */
  cudaThreadSynchronize();
/*-------- stop timing */
  t2 = wall_timer() - t1;
/*--------------------------------------------------*/
  printf("\n=== [GPU] CSR-scalar Kernel ===\n");
  printf("  Number of Threads <%d*%d>\n",gDim,bDim);
  printf("  %.2f ms, %.2f GFLOPS, ",
  t2*1e3,2*nnz/t2/1e9*REPEAT);
/*-------- copy y to host mem */
  cudaMemcpy(y, d_y, n*sizeof(REAL),
  cudaMemcpyDeviceToHost);
/*--------- unbind texture */
  cudaUnbindTexture(texRef);
/*--------- CUDA free */
  cudaFree(d_ia);
  cudaFree(d_ja);
  cudaFree(d_a);
  cudaFree(d_x);
  cudaFree(d_y);
}

__global__ 
void csr_v_k(int n, int *d_ia, int *d_ja, REAL *d_a, REAL *d_y) {
/*------------------------------------------------------------*
 *               CSR spmv-vector kernel
 *  shared memory reduction, texture memory fetching
 *           Half-Warp (16 threads) per row
 *------------------------------------------------------------*/
  // num of half-warps
  int nhw = gridDim.x*BLOCKDIM/HALFWARP;
  // half warp id
  int hwid = (blockIdx.x*BLOCKDIM+threadIdx.x)/HALFWARP;
  // thread lane in each half warp
  int lane = threadIdx.x & (HALFWARP-1);
  // half warp lane in each block
  int hwlane = threadIdx.x/HALFWARP;
  // shared memory for patial result
  volatile __shared__ REAL r[BLOCKDIM+8];
  volatile __shared__ int startend[BLOCKDIM/HALFWARP][2];
  for (int row = hwid; row < n; row += nhw) {
    // row start and end point
    if (lane < 2)
      startend[hwlane][lane] = d_ia[row+lane];
    int p = startend[hwlane][0];
    int q = startend[hwlane][1];
    REAL sum = 0.0;
    for (int i=p+lane; i<q; i+=HALFWARP) {
#if DOUBLEPRECISION
      int2 t = tex1Dfetch(texRef, d_ja[i-1]-1);
      sum += d_a[i-1] * __hiloint2double(t.y, t.x);
#else
      REAL t = tex1Dfetch(texRef, d_ja[i-1]-1);
      sum += d_a[i-1] * t;
#endif
    }
    // parallel reduction
    r[threadIdx.x] = sum;
    r[threadIdx.x] = sum = sum + r[threadIdx.x+8];
    r[threadIdx.x] = sum = sum + r[threadIdx.x+4];
    r[threadIdx.x] = sum = sum + r[threadIdx.x+2];
    r[threadIdx.x] = sum = sum + r[threadIdx.x+1];
    if (lane == 0)
      d_y[row] = r[threadIdx.x];
  }
}

/*-------------------------------------------------------*/
void spmv_csr_vector(struct csr_t *csr, REAL *x, REAL *y) {
  int *d_ia, *d_ja, i;
  REAL *d_a, *d_x, *d_y;
  int n = csr->n;
  int nnz = csr->nnz; 
  double t1,t2;
/*---------- Device Memory */
  cudaMalloc((void **)&d_ia, (n+1)*sizeof(int));
  cudaMalloc((void **)&d_ja, nnz*sizeof(int));
  cudaMalloc((void **)&d_a, nnz*sizeof(REAL));
  cudaMalloc((void **)&d_x, n*sizeof(REAL));
  cudaMalloc((void **)&d_y, n*sizeof(REAL));
/*---------- Memcpy */
  cudaMemcpy(d_ia, csr->ia, (n+1)*sizeof(int),
  cudaMemcpyHostToDevice);
  cudaMemcpy(d_ja, csr->ja, nnz*sizeof(int),
  cudaMemcpyHostToDevice);
  cudaMemcpy(d_a, csr->a, nnz*sizeof(REAL),
  cudaMemcpyHostToDevice);
  cudaMemcpy(d_x, x, n*sizeof(REAL),
  cudaMemcpyHostToDevice);
/*--------- texture binding */
  size_t offset;
#if DOUBLEPRECISION
  cudaBindTexture(&offset, texRef, d_x, n*sizeof(int2));
#else
  cudaBindTexture(&offset, texRef, d_x, n*sizeof(float));
#endif
  assert(offset == 0);
/*-------- set spmv kernel */
/*-------- num of half-warps per block */
  int hwb = BLOCKDIM/HALFWARP;
  int gDim = min(MAXTHREADS/BLOCKDIM, (n+hwb-1)/hwb);
  int bDim = BLOCKDIM;
  //printf("CSR<<<%4d, %3d>>>  ",gDim,bDim);
/*-------- start timing */
  t1 = wall_timer();
  for (i=0; i<REPEAT; i++) {
    //cudaMemset((void *)d_y, 0, n*sizeof(REAL));
    csr_v_k<<<gDim, bDim>>>(n, d_ia, d_ja, d_a, d_y);
  }
/*-------- Barrier for GPU calls */
  cudaThreadSynchronize();
/*-------- stop timing */
  t2 = wall_timer()-t1;
/*--------------------------------------------------*/
  printf("\n=== [GPU] CSR-vector Kernel ===\n");
  printf("  Number of Threads <%d*%d>\n",gDim,bDim);
  printf("  %.2f ms, %.2f GFLOPS, ",
  t2*1e3,2*nnz/t2/1e9*REPEAT);
/*-------- copy y to host mem */
  cudaMemcpy(y, d_y, n*sizeof(REAL),
  cudaMemcpyDeviceToHost);
/*--------- unbind texture */
  cudaUnbindTexture(texRef);
/*---------- CUDA free */
  cudaFree(d_ia);
  cudaFree(d_ja);
  cudaFree(d_a);
  cudaFree(d_x);
  cudaFree(d_y);
}

/*----------------------------------------------*/
/*             THE JAD FORMAT                   */
/*----------------------------------------------*/
__global__ 
void jad_k(int n, int njad, int *d_ia, int *d_ja, 
           REAL *d_a, REAL *d_y, int flag) {
/*-----------------------------------------------*
 *  JAD SpMatVec Kernel(texture memory fetching)
 *            one thread per row
 *-----------------------------------------------*/
  int i,j,p,q;
  REAL r;
/*------------ each thread per row */
  int row = blockIdx.x*blockDim.x+threadIdx.x;
/*------------ number of threads */
  int nthreads = gridDim.x * blockDim.x;
  __shared__ int shia[BLOCKDIM];
  if (threadIdx.x <= njad)
    shia[threadIdx.x] = d_ia[threadIdx.x];
  __syncthreads(); 
  while (row < n) {
    r = 0.0;
    p = shia[0];
    q = shia[1];
    i=0;
    while ( ((p+row) < q) && (i < njad) ) {
      j = p+row;
#if DOUBLEPRECISION
/*--------- double precision texture fetching */
      int2 t = tex1Dfetch(texRef, d_ja[j-1]-1);
      r += d_a[j-1] * __hiloint2double(t.y, t.x);
#else
/*--------- single precision texture fetching */
      r += d_a[j-1] * tex1Dfetch(texRef, d_ja[j-1]-1);
#endif
      i++;
      if (i<njad) {
        p = q;
        q = shia[i+1];
      }
    }
    if (flag)
      d_y[row] += r;
    else
      d_y[row] = r;

    row += nthreads;
  }
}

__global__ 
void vperm_k(int n, REAL *d_y, REAL *d_x, int *d_p) {
/*------------------------------------------------*/
/*   vector permutation, y[p(i)] := x[i]          */
/*------------------------------------------------*/
  int idx = blockIdx.x*blockDim.x+threadIdx.x;
  int nthreads = gridDim.x * blockDim.x;
  int i;
  for (i=idx; i<n; i+=nthreads)
    d_y[d_p[i]-1] = d_x[i];
}

void spmv_jad(struct jad_t *jad, REAL *x, REAL *y) {
  int n = jad->n;
  int nnz = jad->nnz; 
  int *d_ia, *d_ja, *d_p, i, j;
  REAL *d_a, *d_x, *d_y, *d_y2;
  double t1,t2;
/*------------------- allocate Device Memory */
  cudaMalloc((void **)&d_ia, (jad->njad+1)*sizeof(int));
  cudaMalloc((void **)&d_ja, nnz*sizeof(int));
  cudaMalloc((void **)&d_a, nnz*sizeof(REAL));
  cudaMalloc((void **)&d_x, n*sizeof(REAL));
  cudaMalloc((void **)&d_y, n*sizeof(REAL));
  cudaMalloc((void **)&d_y2, n*sizeof(REAL));
  cudaMalloc((void **)&d_p, n*sizeof(int));
/*------------------- Memcpy */
  cudaMemcpy(d_ia, jad->ia, (jad->njad+1)*sizeof(int),
  cudaMemcpyHostToDevice);
  cudaMemcpy(d_ja, jad->ja, nnz*sizeof(int),
  cudaMemcpyHostToDevice);
  cudaMemcpy(d_a, jad->a, nnz*sizeof(REAL),
  cudaMemcpyHostToDevice);
  cudaMemcpy(d_x, x, n*sizeof(REAL),
  cudaMemcpyHostToDevice);
  cudaMemcpy(d_p, jad->perm, n*sizeof(int),
  cudaMemcpyHostToDevice);
/*--------- texture binding */
  size_t offset;
#if DOUBLEPRECISION
  cudaBindTexture(&offset, texRef, d_x, n*sizeof(int2));
#else
  cudaBindTexture(&offset, texRef, d_x, n*sizeof(float));
#endif
  assert(offset == 0);
/*--------- set spmv kernel */
  int nthreads = min(MAXTHREADS, n);
  int gDim = (nthreads+BLOCKDIM-1)/BLOCKDIM;
  int bDim = BLOCKDIM;
/*-------- start timing */
  t1 = wall_timer();
  for (i=0; i<REPEAT; i++) {
    //cudaMemset((void *)d_y, 0, n*sizeof(REAL));
/*-------- handle (BLOCKDIM-1) jads each time */
    for (j=0; j<jad->njad; j+=(BLOCKDIM-1)) {
      int njad = min((jad->njad-j), (BLOCKDIM-1));
      jad_k<<<gDim, bDim>>>
      (n, njad, d_ia+j, d_ja, d_a, d_y,j);
    }
  }
/*-------- permutation */
  vperm_k<<<gDim, bDim>>>(n, d_y2, d_y, d_p);
/*-------- barrier for GPU calls */
  cudaThreadSynchronize();
/*-------- stop timing */
  t2 = wall_timer()-t1;
/*--------------------------------------------------*/
  printf("\n=== [GPU] JAD Kernel ===\n");
  printf("  NJAD %d\n", jad->njad);
  printf("  Number of Threads <%d*%d>\n",gDim,bDim);
  printf("  %.2f ms, %.2f GFLOPS, ",
  t2*1e3,2*nnz/t2/1e9*REPEAT);
/*-------- copy y to host mem */
  cudaMemcpy(y, d_y2, n*sizeof(REAL), 
  cudaMemcpyDeviceToHost);
/*--------- unbind texture */
  cudaUnbindTexture(texRef);
/*--------- CUDA free */
  cudaFree(d_ia);
  cudaFree(d_ja);
  cudaFree(d_a);
  cudaFree(d_x);
  cudaFree(d_y);
  cudaFree(d_y2);
  cudaFree(d_p);
}

/*---------------------------------------------------*/
/*                THE DIA FORMAT                     */
/*---------------------------------------------------*/
__global__ 
void dia_k(int n, int stride, int ndiags, int *d_ioff, 
              REAL *d_diags, REAL *d_y, int flag) {
/*---------------------------------------------------*
 *    DIA SpMatVec Kernel(texture memory fetching)
 *               one thread per row
 *---------------------------------------------------*/
/*--------- number of threads */
  int nthreads = gridDim.x * blockDim.x;
  __shared__ int shioff[BLOCKDIM];
  if (threadIdx.x < ndiags)
    shioff[threadIdx.x] = d_ioff[threadIdx.x];
  __syncthreads();
/*--------- each thread per row */
  int row = blockIdx.x*blockDim.x+threadIdx.x;
/*--------- start position of a row */
  d_diags += row;
  while (row < n) {
    REAL r = 0;
    for (int i=0; i<ndiags; i++) {
/*--------- col index */
      int q = shioff[i] + row;
      if (q >= 0 && q < n) {
#if DOUBLEPRECISION
/*--------- double precision texture fetching */
        int2 t = tex1Dfetch(texRef, q);
        r += *(d_diags + i*stride) * __hiloint2double(t.y, t.x);
#else
/*--------- single precision texture fetching */
        r += *(d_diags + i*stride) * tex1Dfetch(texRef, q);
#endif
      }
    }
    if (flag)
      d_y[row] += r;
    else
      d_y[row] = r;
    row += nthreads;
    d_diags += nthreads;
  }
}

/*-------------------------------------------------*/
void spmv_dia(struct dia_t *dia, REAL *x, REAL *y) {
/*-------- GPU DIA format SpMV */
  int n = dia->n;
  int *d_ioff, i, j;
  REAL *d_diags, *d_x, *d_y;
  double t1,t2;
/*--------- Device Memory */
  cudaMalloc((void **)&d_diags, 
  dia->ndiags*dia->stride*sizeof(REAL));
  cudaMalloc((void **)&d_ioff, 
  dia->ndiags*sizeof(int));
  cudaMalloc((void **)&d_x, (size_t)n*sizeof(REAL));
  cudaMalloc((void **)&d_y, (size_t)n*sizeof(REAL));
/*--------- Memcpy */
  cudaMemcpy(d_diags, dia->diags,
  dia->ndiags*dia->stride*sizeof(REAL),
  cudaMemcpyHostToDevice);
  cudaMemcpy(d_ioff, dia->ioff,
  dia->ndiags*sizeof(int),
  cudaMemcpyHostToDevice);
  cudaMemcpy(d_x, x, (size_t)n*sizeof(REAL), 
  cudaMemcpyHostToDevice);
/*--------- texture binding */
  size_t offset;
#if DOUBLEPRECISION
  cudaBindTexture(&offset, texRef, d_x, n*sizeof(int2));
#else
  cudaBindTexture(&offset, texRef, d_x, n*sizeof(float));
#endif
  assert(offset == 0);
/*------- set spmv kernel */
  int nthreads = min(MAXTHREADS, n);
  int gDim = (nthreads+BLOCKDIM-1)/BLOCKDIM;
  int bDim = BLOCKDIM;
/*-------- start timing */
  t1 = wall_timer();
  for (i=0; i<REPEAT; i++) {
    //cudaMemset((void *)d_y, 0, n*sizeof(REAL));
/*-------- handle BLOCKDIM diags each time */
    for (j=0; j<dia->ndiags; j+=BLOCKDIM) {
      int ndiags = min((dia->ndiags-j), BLOCKDIM);
      dia_k<<<gDim, bDim>>>
      (n,dia->stride,ndiags,d_ioff+j,d_diags+j*n,d_y,j);
    }
  }
/*-------- barrier for GPU calls */
  cudaThreadSynchronize();
/*-------- stop timing */
  t2 = wall_timer()-t1;
/*--------------------------------------------------*/
  printf("\n=== [GPU] DIA-vector Kernel ===\n");
  printf("  Padding with Zeros %.2f\n",\
  double(dia->ndiags*n)/(double)dia->nnz);
  printf("  NDIAGS %d\n", dia->ndiags);
  printf("  Number of Threads <%d*%d>\n",gDim,bDim);
  printf("  %.2f ms, %.2f GFLOPS, ",
  t2*1e3, 2*(dia->nnz)/t2/1e9*REPEAT);
/*-------- copy y to host mem */
  cudaMemcpy(y, d_y, (size_t)n*sizeof(REAL),
  cudaMemcpyDeviceToHost);
/*--------- unbind texture */
  cudaUnbindTexture(texRef);
/*--------- CUDA free */
  cudaFree(d_diags);
  cudaFree(d_ioff);
  cudaFree(d_x);
  cudaFree(d_y);
}

/*----------------------------------------------------*/
void spmv_csr_cpu(struct csr_t *csr, REAL *x, REAL *y) {
/*------------- CPU CSR SpMV kernel */
  int ii,i,j;
  REAL r;
  double t1, t2;
  t1 = wall_timer();
  for (ii=0; ii<REPEAT; ii++) {
    //memset(y, 0, csr->n*sizeof(REAL));  
    for (i=0; i<csr->n; i++) {
      r = 0.0;
      for (j=csr->ia[i]; j<csr->ia[i+1]; j++)
        r += csr->a[j-1]*x[csr->ja[j-1]-1];
      y[i] = r;
    }
  }
  t2 = wall_timer() - t1;
/*--------------------------------------------------*/
  printf("\n=== [CPU] CSR Kernel ===\n");
  printf("  %.2f ms, %.2f GFLOPS ",
  t2*1e3, 2*(csr->nnz)/t2/1e9*REPEAT);
}

/*-----------------------------------------------*/
void cuda_init(int argc, char **argv) {
  int deviceCount, dev;
  cudaGetDeviceCount(&deviceCount);
  printf("=========================================\n");
  if (deviceCount == 0)
    printf("There is no device supporting CUDA\n");

  for (dev = 0; dev < deviceCount; ++dev) {
    cudaDeviceProp deviceProp;
    cudaGetDeviceProperties(&deviceProp, dev);
    if (dev == 0) {
      if (deviceProp.major == 9999 && deviceProp.minor == 9999)
        printf("There is no device supporting CUDA.\n");
      else if (deviceCount == 1)
        printf("There is 1 device supporting CUDA\n");
      else
        printf("There are %d devices supporting CUDA\n", deviceCount);
    }
    printf("\nDevice %d: \"%s\"\n", dev, deviceProp.name);
    printf("  Major revision number:          %d\n",
           deviceProp.major);
    printf("  Minor revision number:          %d\n",
           deviceProp.minor);
    printf("  Total amount of global memory:  %.2f GB\n",
           deviceProp.totalGlobalMem/1e9);
  }
  dev = 0;
  cudaSetDevice(dev);
  cudaDeviceProp deviceProp;
  cudaGetDeviceProperties(&deviceProp, dev);
  printf("\nRunning on Device %d: \"%s\"\n", dev, deviceProp.name);
  printf("=========================================\n");
}

/*---------------------------------------------------*/
void cuda_check_err() {
  cudaError_t cudaerr = cudaGetLastError() ;
  if (cudaerr != cudaSuccess) 
    printf("error: %s\n",cudaGetErrorString(cudaerr));
}

